import numpy as np

from irmetrics.io import _ensure_io, _validate_unique
from irmetrics.relevance import multilabel


@_ensure_io
@_validate_unique
def rr(y_true, y_pred, k=None, relevance=multilabel):
    """Compute Recirocal Rank(s).
    Calculate the recirocal of the index for the first matched item in
    ``y_pred``. The score is between 0 and 1.

    This ranking metric yields a high value if true labels are ranked high by
    ``y_pred``.

    Parameters
    ----------
    y_true : scalar, iterable or ndarray of shape (n_samples, n_labels)
        True labels of entities to be ranked. In case of scalars ``y_pred``
        should be of shape (1, n_labels).
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.

    Returns
    -------
    rr : float in [0., 1.]
        The recirocal ranks for all samples.

    References
    ----------
    `Wikipedia entry for Mean reciprocal rank
    <https://en.wikipedia.org/wiki/Mean_reciprocal_rank>`_

    Examples
    --------
    >>> from irmetrics.topk import rr
    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [0, 1, 4]
    >>> rr(y_true, y_pred)
    0.5
    """
    relevant = relevance(y_true, y_pred)
    index = relevant.argmax(-1)
    return relevant.any(-1) / (index + 1)


@_ensure_io
@_validate_unique
def recall(y_true, y_pred=None, k=None, relevance=multilabel, pad_token=None):
    """Compute Recall(s).
    Check if at least one metric proposed in ``y_pred`` is in ``y_true``.
    This is the binary score, 0 -- all predictionss are irrelevant
    and 1 otherwise.
    This definition of recall is equivalent to accuracy@k.

    Parameters
    ----------
    y_true : scalar, iterable or ndarray of shape (n_samples, n_labels)
        True labels of entities to be ranked. In case of scalars ``y_pred``
        should be of shape (1, n_labels).
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.
    pad_token : callable, default=None
        A value that was used to pad the `y_true`. This is needed to ignore
        the padding when calculating the recall. The default value is `None`,
        this means `numpy` will do element-wise comparison for `None`, change
        it to somehting else if your predictions may contain `None` objects.

    Returns
    -------
    recall : float in [0., 1.]
        The relevances for all samples.

    References
    ----------
    `Wikipedia entry for precision and recall
    <https://en.wikipedia.org/wiki/Precision_and_recall>`_

    Examples
    --------
    >>> from irmetrics.topk import recall

    for ground-truth labels related to a query:

    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [0, 1, 4]
    >>> recall(y_true, y_pred)
    1.0
    """
    # This performes element-wise comparison if dtypes agree
    positives = ~(y_true == pad_token)
    return relevance(y_true, y_pred).sum(-1) / positives.sum(-1)


@_ensure_io
@_validate_unique
def precision(y_true, y_pred=None, k=None, relevance=multilabel):
    """Compute Recall(s).
    and 1 otherwise.
    Check which fraction of ``y_pred`` is in ``y_true``.
    **NB**: When passing ``y_pred`` of shape `[n_samples, n_outputs]`
    the result is quivalent to `recall(y_pred, y_true) / n_outputs`.

    Parameters
    ----------
    y_true : scalar, iterable or ndarray of shape (n_samples, n_labels)
        True labels of entities to be ranked. In case of scalars ``y_pred``
        should be of shape (1, n_labels).
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.

    Returns
    -------
    precision : float in [0., 1.]
        The relevances for all samples.

    References
    ----------
    `Wikipedia entry for precision and recall
    <https://en.wikipedia.org/wiki/Precision_and_recall>`_

    Examples
    --------
    >>> from irmetrics.topk import recall

    for ground-truth labels related to a query:

    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [0, 1, 4, 3]
    >>> precision(y_true, y_pred)
    0.25
    """
    return relevance(y_true, y_pred).sum(-1) / y_pred.shape[-1]


def dcg_score(relevance, k=None, weights=1.0):
    """Compute Discounted Cumulative Gain score(s) based on `relevance`
    judgements provided.

    This is provided as internal implementation for `ndcg` for this reason
    the API for this function slightly differ: it alawyas accepts  and
    outputs `np.arrays`, unlike other methos in this module.

    Parameters
    ----------
    relevance : iterable or ndarray of shape (n_samples, n_labels) or simply
        (n_labels,). The last dimension of the parameter is used as position.
        The relevance judgements provided by experts.
    weights : default=1.0, scalar, iterable or ndarray of shape (n_samples,)
        takes into account the importance of each sample, if relevant.
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.

    Returns
    -------
    dcg : np.array
        The discounted cumulative gains for samples (or a single sample).

    References
    ----------
    `Wikipedia entry for Discounted cumulative gain
    <https://en.wikipedia.org/wiki/Discounted_cumulative_gain#Discounted_Cumulative_Gain>`_

    Examples
    --------
    >>> from irmetrics.topk import dcg_score

    for ground-truth labels related to a query:

    >>> relevance_judgements = np.array([[1, 0, 0, 0]])
    >>> dcg_score(relevance_judgements)
    array([1.])
    >>> relevance_judgements = np.array([[True, False, False, False]])
    >>> dcg_score(relevance_judgements)
    array([1.])
    >>> relevance_judgements = np.array([[False, True, False, False]])
    >>> dcg_score(relevance_judgements)
    array([0.63092975])
    """
    top = relevance[..., :k]
    gains = (2 ** top - 1) / np.log2(np.arange(top.shape[-1]) + 2)[None, ...]
    return np.sum(gains * weights, axis=-1)


@_ensure_io
@_validate_unique
def ndcg(y_true, y_pred, k=None, relevance=multilabel, weights=1.):
    """Compute Normalized Discounted Cumulative Gain score(s) based on
    `relevance` judgements provided.

    Parameters
    ----------
    y_true : iterable or ndarray of shape (n_samples, n_labels) or simply
        (n_labels,). The last dimension of the parameter is used as position.
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.
    weights : float, iterable, ndarray, default=1.0
        Represents the weights of each sample.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.

    Returns
    -------
    ndcg : np.array
        The discounted cumulative gains for samples (or a single sample).

    References
    ----------
    `Wikipedia entry for normalized discounted cumulative gain
    <https://en.wikipedia.org/wiki/Discounted_cumulative_gain#Normalized_DCG>`_

    Examples
    --------
    >>> from irmetrics.topk import ndcg

    for ground-truth labels related to a query:

    >>> y_true = [1, 2]
    >>> y_pred = [0, 1, 0, 0]
    >>> ndcg(y_true, y_pred)
    0.6309297535714575
    >>> # the order of y_true labels doesn't matter
    >>> y_true = [2, 1]
    >>> y_pred = [0, 1, 0, 0]
    >>> ndcg(y_true, y_pred)
    0.6309297535714575
    """
    relevant = relevance(y_true, y_pred)

    # Sort in descending order, calculate the gain
    idcg = dcg_score(np.flip(np.sort(relevant, axis=-1), axis=-1), k, weights)

    # Normalize to the ideal dcg score
    return dcg_score(relevant, k) / idcg


@_ensure_io
@_validate_unique
def ap(y_true, y_pred, k=None, relevance=multilabel):
    """Compute Average Precision score(s).
    AP is an aproximation of the integral over PR-curve.

    Parameters
    ----------
    y_true : scalar, iterable or ndarray of shape (n_samples, n_labels)
        True labels of entities to be ranked. In case of scalars ``y_pred``
        should be of shape (1, n_labels).
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs. The minimum between the nuber of correct answers and k will
        be used to compute the score.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.

    Returns
    -------
    ap : float
        The average precision for a given sample.

    References
    ----------
    `Wikipedia entry for Mean Average Precision
    <https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision>`_

    Examples
    --------
    >>> from irmetrics.topk import ap

    for ground-truth labels related to a query:

    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [1, 0, 0]
    >>> ap(y_true, y_pred)
    0.3333333333333333
    >>> # This should be fixed
    >>> y_true = [1, 4, 5]

    and the predicted labels by an IR system:

    >>> y_pred = [1, 2, 3, 4, 5]
    >>> ap(y_true, y_pred)
    array([0.2, 0. , 0. ])
    """
    relevant = relevance(y_true, y_pred)

    # Handle k=None, without if else branching
    max_iter = min(i for i in (k, y_true.shape[-1]) if i is not None)

    ap = np.sum([
        np.array(
            precision(y_true, y_pred, ik + 1)
        )[..., None] * relevant[..., [ik]]
        for ik in range(max_iter)
    ], axis=-1)

    return ap / y_pred.shape[-1]



from pallets_sphinx_themes import get_version
from pallets_sphinx_themes import ProjectLink

# Project
project = "ir-metrics"
copyright = "2020 kqf"
author = "kqf"
release, version = get_version("ir-metrics", version_length=1)

# General
master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.log_cabinet",
    "pallets_sphinx_themes",
    "sphinx_issues",
    "sphinx_tabs.tabs",
    "numpydoc",
]
intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}
issues_github_path = "kqf/ir-metrics"

# HTML
html_theme = "flask"
html_theme_options = {"index_sidebar_logo": False}
html_context = {
    "project_links": [
        ProjectLink("Website", "https://ir-metrics.readthedocs.io"),
        ProjectLink("PyPI releases", "https://pypi.org/project/ir-metrics/"),
        ProjectLink("Source Code", "https://github.com/kqf/ir-metrics/"),
        ProjectLink("Issues", "https://github.com/kqf/ir-metrics/issues/"),
    ]
}
html_sidebars = {
    "index": ["project.html", "localtoc.html", "searchbox.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html"],
}
singlehtml_sidebars = {"index": ["project.html", "localtoc.html"]}
# html_static_path = ["_static"]
# html_favicon = "_static/ir-metrics-icon.png"
# html_logo = "_static/ir-metrics-logo-sidebar.png"
html_title = f"ir-metrics Documentation ({version})"
html_show_sourcelink = False

# LaTeX--
latex_documents = [(
    master_doc, f"irm-{version}.tex", html_title, author, "manual")]



import pytest
import numpy as np

from contextlib import contextmanager
from irmetrics.coverage import coverage, iou


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize("y_pred, output", [
    ([1, 0, 0], 1),
    ([0, 1, 0], 1.),
    ([0, 0, 1], 1.),
    ([1, 1, 1], 1.),
    ([0, 0, 0], 1.),
    ([0, 0, 0] * 20 + [1], 1),
    ([0, None], 1),
    ([None], 0),
])
def test_coverage(y_pred, output, n_samples=128):
    assert coverage(y_pred) == output

    # Now the vectorized output
    y_preds = np.repeat(np.atleast_2d(y_pred), n_samples, axis=0)
    outputs = np.repeat(np.array(output), n_samples)
    assert np.all(coverage(y_preds) == outputs)


@pytest.mark.parametrize(
    "y_true, y_pred, output, exception",
    [
        (1, [1, 2, 3], 1. / 3, does_not_raise()),
        (1, [2, 1, 3], 1. / 3, does_not_raise()),
        (1, [2, 3, 1], 1. / 3, does_not_raise()),
        (1, [3, 4, 5], 0, does_not_raise()),
        (1, [1, 1, 1], 0, pytest.raises(ValueError)),
        ([1, 1, 1], 1, 0, pytest.raises(ValueError)),
    ])
def test_iou(y_true, y_pred, output, exception, n_samples=128):
    with exception:
        assert iou(y_true, y_pred) == output


@pytest.mark.parametrize("y_true, y_pred, output, exception", [
    (1, [1, 2, 3], 1. / 3, does_not_raise()),
    (1, [2, 1, 3], 1. / 3, does_not_raise()),
    (1, [2, 3, 1], 1. / 3, does_not_raise()),
    (1, [3, 4, 5], 0, does_not_raise()),
    (1, [1, 1, 1], 0, pytest.raises(ValueError)),
    ([1, 1, 1], 1, 0, pytest.raises(ValueError)),
])
def test_iou_vectorized(y_true, y_pred, output, exception, n_samples=128):
    y_trues = np.repeat(np.array(y_true), n_samples)
    y_preds = np.repeat(np.atleast_2d(y_pred), n_samples, axis=0)
    outputs = np.repeat(np.array(output), n_samples)
    with exception:
        np.testing.assert_array_equal(iou(y_trues, y_preds), outputs)



import pytest
import numpy as np

from contextlib import contextmanager
from irmetrics.topk import rr, recall, precision, ndcg, ap


@contextmanager
def does_not_raise():
    yield


@pytest.fixture
def inputs():
    return [
        (1, [1, 2, 3]),
        (1, [2, 1, 3]),
        (1, [2, 3, 1]),
        (1, [1, 2, 3]),
        (1, [np.nan, np.nan, np.nan]),
        (1, [8, 9, 10]),  # all irrelevant
        ("item1", ["item1", "item2", "item3"]),
        ("item1", ["item1", np.nan, np.nan]),  # noqa, raises as nans are converted to str
        # (1, [1, None, None]),
    ]


_OUTPUTS = {
    rr: [
        (1, does_not_raise),
        (1. / 2, does_not_raise),
        (1. / 3, does_not_raise),
        (1., does_not_raise),
        (0, does_not_raise),
        (0, does_not_raise),
        (1, does_not_raise),
        (1, does_not_raise),
        # (1, does_not_raise),
    ],
    recall: [
        (1, does_not_raise),
        (1, does_not_raise),
        (1, does_not_raise),
        (1, does_not_raise),
        (0, does_not_raise),
        (0, does_not_raise),
        (1, does_not_raise),
        (1, does_not_raise),
        # (1, does_not_raise),
    ],
    precision: [
        (1. / 3., does_not_raise),
        (1. / 3., does_not_raise),
        (1. / 3., does_not_raise),
        (1. / 3., does_not_raise),
        (0., does_not_raise),
        (0., does_not_raise),
        (1. / 3., does_not_raise),
        (1. / 3., does_not_raise),
        # (1. / 3., does_not_raise),
    ],
    ndcg: [
        (1., does_not_raise),
        (1. / np.log2(2 + 1), does_not_raise),  # idx + 1, (idx starts from 1)
        (1. / np.log2(3 + 1), does_not_raise),  # idx + 1, (idx starts from 1)
        (1., does_not_raise),
        (np.nan, does_not_raise),
        (np.nan, does_not_raise),
        (1., does_not_raise),
        (1., does_not_raise),
        # (1., does_not_raise),
    ],
    ap: [
        (1. / 3., does_not_raise),
        (0., does_not_raise),
        (0., does_not_raise),
        (1. / 3., does_not_raise),
        (0., does_not_raise),
        (0., does_not_raise),
        (1. / 3., does_not_raise),
        (1. / 3., does_not_raise),
        # (1. / 3., does_not_raise),
    ],
}


@pytest.fixture
def raw_outputs(measure):
    return list(zip(*_OUTPUTS[measure]))


@pytest.fixture
def expected(raw_outputs):
    return raw_outputs[0]


@pytest.fixture
def exceptions(raw_outputs):
    return raw_outputs[1]



def unilabel(y_true, y_pred):
    """Compute relevance(s) of predicted labels.
    This version of the relevance function works only for the queries
    (problems) with a single groud truth label.

    It is provided mainly for two reasons: there is a slight speedup (order of
    seconds for the large `n_samples`) and it adds expresivity
    if needed.

    Parameters
    ----------
    y_true : ndarray of shape (n_samples, 1), where `n_samples >= 1`
        Ground true labels for a given query (as returned by an IR system).
    y_pred : ndarray of shape (n_samples, n_labels), where `n_samples >= 1`
        Target labels sorted by relevance (as returned by an IR system).

    Returns
    -------
    relevance : bolean ndarray
        The relevance judgements for `y_pred` of shape (n_samples, 1)

    Raises
    -------
    ValueError
        If `y_true` has last dimension larger than 1 (multilabel case).

    Examples
    --------
    >>> import numpy as np
    >>> from irmetrics.relevance import unilabel
    >>> # ground-truth label of some answers to a query:
    >>> y_true = np.array([[1]]) # (1, 1)

    and the predicted labels by an IR system:

    >>> y_pred = np.array([[0, 1, 4]]) # (1, 3)
    >>> unilabel(y_true, y_pred)
    array([[False,  True, False]])
    >>> y_true = np.array([[1], [2]]) # (2, 1)
    >>> y_pred = np.array([[0, 1, 4], [5, 6, 7]]) # (2, 3)
    >>> unilabel(y_true, y_pred)
    array([[False,  True, False],
           [False, False, False]])
    """
    if y_true.shape[-1] != 1:
        msg = "y_true is expected to be of shape (n_samples, 1), got {}"
        raise ValueError(msg.format(y_true.shape))
    return y_true == y_pred


def multilabel(y_true, y_pred):
    """Compute relevance(s) of predicted labels.

    Parameters
    ----------
    y_true : ndarray of shape (n_samples, n_true), where `n_samples >= 1`
        Ground true labels for a given query (as returned by an IR system).
    y_pred : ndarray of shape (n_samples, n_labels), where `n_samples >= 1`
        Target labels sorted by relevance (as returned by an IR system).
        The `n_labels` and `n_true` may not be the same.

    Returns
    -------
    relevance : bolean ndarray
        The relevance judgements for `y_pred` of shape (n_samples, n_labels)

    Examples
    --------
    >>> import numpy as np
    >>> from irmetrics.relevance import multilabel
    >>> # ground-truth label of some answers to a query:
    >>> y_true = np.array([[1]]) # (1, 1)

    and the predicted labels by an IR system:

    >>> y_pred = np.array([[0, 1, 4]]) # (1, 3)
    >>> multilabel(y_true, y_pred)
    array([[False,  True, False]])
    >>> y_true = np.array([[1], [2]]) # (2, 1)
    >>> y_pred = np.array([[0, 1, 4], [5, 6, 7]]) # (2, 3)
    >>> multilabel(y_true, y_pred)
    array([[False,  True, False],
           [False, False, False]])
    >>> # Now the multilabel case:
    >>> y_true = np.array([[1, 4]]) # (1, 2)
    >>> y_pred = np.array([[0, 1, 4]]) # (1, 3)
    >>> multilabel(y_true, y_pred)
    array([[False,  True,  True]])
    """
    return (y_pred[:, :, None] == y_true[:, None]).any(axis=-1)


def relevant_counts(y_pred, y_true):
    """Calculate the total number of relevant items.

    Parameters
    ----------
    y_true : ndarray of shape (n_samples, n_true), where `n_samples >= 1`
        Ground true labels for a given query (as returned by an IR system).
    y_pred : ndarray of shape (n_samples, n_labels), where `n_samples >= 1`
        Target labels sorted by relevance (as returned by an IR system).
        The `n_labels` and `n_true` may not be the same.

    Returns
    -------
    relevance_counts: ndarray
        The number of true relevance judgements for `y_pred`.

    Examples
    --------
    >>> import numpy as np
    >>> from irmetrics.relevance import relevant_counts
    >>> # ground-truth label of some answers to a query:
    >>> y_true = np.array([[1]]) # (1, 1)

    and the predicted labels by an IR system:

    >>> y_pred = np.array([[0, 1, 4]]) # (1, 3)
    >>> relevant_counts(y_true, y_pred)
    array([[1]])
    >>> y_true = np.array([[1], [2]]) # (2, 1)
    >>> y_pred = np.array([[0, 1, 4], [5, 6, 7]]) # (2, 3)
    >>> relevant_counts(y_true, y_pred)
    array([[1],
           [1]])
    >>> # Now the `relevant_counts` case:
    >>> y_true = np.array([[1, 4]]) # (1, 2)
    >>> y_pred = np.array([[0, 1, 4]]) # (1, 3)
    >>> relevant_counts(y_true, y_pred)
    array([[1, 1]])
    """
    return (y_pred[:, :, None] == y_pred[:, None]).sum(axis=-1)



import pytest
import numpy as np
import pandas as pd
from irmetrics.flat import flat
from irmetrics.topk import rr, ndcg


@pytest.fixture
def data(inputs):
    # Explode the examples
    df = pd.DataFrame([
        {
            "query": i,
            "relevance": answer == prediction,
            "weights": 1,
        }
        for i, (answer, predictions) in enumerate(inputs)
        for prediction in predictions
    ])
    return df


@pytest.mark.parametrize("measure", [
    rr,
    ndcg,
    # Not going to support these methods as they require
    # true shapes of y_pred/y_true.
    # recall,
    # precision,
    # ap,
])
def test_calculates_data(data, expected, measure):
    outputs = flat(
        data,
        query_col="query",
        relevance_col="relevance",
        measure=measure,
    )
    np.testing.assert_almost_equal(outputs.values, expected)


"""
    This is a separate test for NDCG as it might accept nonbinary judgements
"""

EXAMPLES = [
    ((3, 2, 3, 0, 1, 2), 0.948810748),
    ((1, 0, 0, 0, 0, 0), 1.0),
    ((0, 1, 0, 0, 0, 0), 0.630929753),
    ((1, 1, 0, 0, 0, 0), 1.0),
    ((0, 0, 0, 0, 0, 1), 0.356207187),
    ((0, 0, 0, 0, 0, 0), np.NaN)
]


@pytest.fixture
def nonbinary_data():
    queries, expected = zip(*EXAMPLES)
    # Explode the examples
    df = pd.DataFrame([
        {
            "query": i,
            "relevance": rel,
            "weights": 1,
        }
        for i, relevance in enumerate(queries)
        for rel in relevance
    ])
    return df, expected


def test_nonbinary_ndcg(nonbinary_data):
    df, expected = nonbinary_data
    outputs = flat(
        df,
        query_col="query",
        relevance_col="relevance",
        measure=ndcg,
    )
    np.testing.assert_almost_equal(outputs.values, expected)



import numpy as np
from irmetrics.io import to_scalar, _ensure_io
from irmetrics.relevance import multilabel, relevant_counts


def coverage(y_pred, pad_token=None):
    """Compute Coverage(s)
    Check if ``y_pred`` contains any nontrivial results.

    Parameters
    ----------
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    pad_token : scalar, str, default=None
        The value that was used to pad the predictions to get the same length.

    Returns
    -------
    coverage : int in [0, 1]
        The coverage is 1 if ``y_pred`` contains any results different from
                ``pad_token`` and 0 otherwise.

    Examples
    --------
    >>> from irmetrics.topk import rr

    for gound-truth labels related to some query

    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [0, 1, 4]
    >>> coverage(y_true)
    1
    >>> y_pred = [0, None]
    >>> coverage(y_true)
    1
    >>> coverage([-1], pad_token=-1)
    0
    """
    outputs = np.not_equal(y_pred, pad_token).sum(axis=-1) > 0
    return to_scalar(outputs.astype(np.int32))


@_ensure_io
def iou(y_true, y_pred, k=None, relevance=multilabel, n_uniq=relevant_counts):
    """Compute the approximate version of Intersection over Union.
    The approximation comes in assumption that `y_true` and `y_pred`
    contain only unique values.

    Parameters
    ----------
    y_true : scalar, iterable or ndarray of shape (n_samples, n_labels)
        True labels of entities to be ranked. In case of scalars ``y_pred``
        should be of shape (1, n_labels).
    y_pred : iterable, ndarray of shape (n_samples, n_labels)
        Target labels sorted by relevance (as returned by an IR system).
    k : int, default=None
        Has no effect provided only for api compatibility.
    relevance : callable, default=topk.relevance.multilabel
        A function that calculates relevance judgements based on input
        ``y_pred`` and ``y_true``.
    n_uniq : callable, default=topk.relevance.relevant_counts
        A function that calculates number of unique labels per query.

    Returns
    -------
    iou : float in [0., 1.]
        The ratio of relevant retrieved entries to the union of relevant
        and retrieved entries.

    References
    ----------
    `Wikipedia entry for Jaccard Index
    <https://en.wikipedia.org/wiki/Jaccard_index>`_

    Examples
    --------
    >>> from irmetrics.topk import rr

    for ground-truth labels related to a query:

    >>> y_true = 1

    and the predicted labels by an IR system:

    >>> y_pred = [0, 1, 4]
    >>> iou(y_true, y_pred)
    0.3333333333333333
    """

    if np.any(n_uniq(y_pred, y_true) > 1):
        raise ValueError("y_pred has duplicates along the last axis")

    if np.any(n_uniq(y_true, y_true) > 1):
        raise ValueError("y_true has duplicates along the last axis")

    relevant = relevance(y_pred, y_true)

    # Approximate intersection
    intersection = relevant.sum(axis=-1)

    # Approximate union
    union = (y_pred.shape[-1] + y_true.shape[-1] - intersection)

    return intersection / union



import pytest
import numpy as np

from irmetrics.topk import rr, recall, precision, ndcg, ap
from irmetrics.relevance import unilabel, multilabel


@pytest.fixture
def cases(inputs, expected, exceptions):
    return zip(inputs, expected, exceptions)


@pytest.mark.parametrize("measure", [
    rr,
    recall,
    precision,
    ndcg,
    ap,
])
@pytest.mark.parametrize("relevance", [
    unilabel,
    multilabel,
])
def test_all(cases, measure, relevance):
    for (y_true, y_pred), expected, exception in cases:
        with exception():
            np.testing.assert_equal(
                measure(y_true, y_pred, relevance=relevance),
                expected
            )


@pytest.mark.parametrize("measure", [
    rr,
    recall,
    precision,
    ndcg,
    ap,
])
@pytest.mark.parametrize("relevance", [
    unilabel,
    multilabel,
])
def test_all_vectorized(cases, measure, relevance, n_samples=128):
    for (y_true, y_pred), expected, exception in cases:

        # Repeat the inputs and expected values along the batch dimension
        y_trues = np.tile(np.array(y_true), (n_samples, 1))
        y_preds = np.tile(np.atleast_2d(y_pred), (n_samples, 1))
        outputs = np.repeat(np.array(expected), n_samples)

        with exception():
            np.testing.assert_equal(
                measure(y_trues, y_preds, relevance=relevance),
                outputs
            )


@pytest.mark.parametrize("measure", [recall])
@pytest.mark.parametrize("relevance", [
    multilabel,
])
@pytest.mark.parametrize("pad_token", [
    None,
])
def test_recall_padding(cases, measure, relevance, pad_token):
    for (y_true, y_pred), expected, exception in cases:
        y_true = [y_true, pad_token]
        with exception():
            np.testing.assert_equal(
                measure(
                    y_true,
                    y_pred,
                    relevance=relevance,
                    pad_token=pad_token
                ),
                expected
            )



import pytest
import numpy as np

from irmetrics.relevance import unilabel, multilabel
from contextlib import contextmanager


def arr(x):
    return np.array([x])


@pytest.mark.parametrize("y_true, y_pred, output", [
    (arr([1]), arr([1, 2, 3]), arr([True, False, False])),
    (arr([1]), arr([2, 2, 3]), arr([False, False, False])),
])
@pytest.mark.parametrize("relevance", [unilabel, multilabel])
def test_single_labels(y_true, y_pred, output, relevance):
    np.testing.assert_equal(relevance(y_true, y_pred), output)


@contextmanager
def does_not_raise():
    yield


def conditional_raises(relevance):
    if relevance == unilabel:
        return pytest.raises(ValueError)
    return does_not_raise()


@pytest.mark.parametrize("y_true, y_pred, output", [
    (arr([1, 2]), arr([1, 2, 3]), arr([True, True, False])),
    (arr([1, 2]), arr([2, 2, 3]), arr([True, True, False])),
])
@pytest.mark.parametrize("relevance", [unilabel, multilabel])
def test_multiple_labels(y_true, y_pred, output, relevance):
    with conditional_raises(relevance):
        np.testing.assert_equal(relevance(y_true, y_pred), output)



import pytest
import numpy as np

from numpy import array as ar
from irmetrics.io import ensure_inputs


# Identity shortcut
_id = ar([[1]])
_K = 20


@pytest.mark.parametrize("y_pred, y_pred_ex", [
    (1, _id),  # scalar -> [1, 1]
    ([1], _id),  # [1] -> [1, 1]
    ([1, 2, 3, 4], ar([[1, 2, 3, 4]])),  # [n] -> [1, n]
    ([[1]], _id),  # [1, 1] -> [1, 1]
    ([[1, 2]], ar([[1, 2]])),  # [1, 2] -> [1, 2]
    ([[1], [2]], ar([[1], [2]])),  # [2, 1] -> [2, 1]
    # Realistic case: [n_samples, n_preds] -> [n_samples, k]
    (np.tile(1, (128, 40)), np.tile(1, (128, _K))),
])
@pytest.mark.parametrize("y_true, y_true_ex", [
    (1, _id),  # scalar -> [1, 1]
    ([1], _id),  # [1] -> [1, 1]
    ([[1]], _id),  # [1, 1] -> [1, 1]
    # Realistic case: [n_samples, n_preds] -> [n_samples, k]
    (np.tile(1, (128, 40)), np.tile(1, (128, _K))),
    # The cases below depend on the y_pred input
    # ([[1, 2]], ar([[1, 2]])),  # [1, 2] -> [1, 2]
    # ([[1], [2]], ar([[1], [2]])),  # [2, 1] -> [2, 1]
    # ([1, 2, 3, 4], ar([[1], [2], [3], [4]])),  # [n] -> [n, 1]
])
def test_handles_inputs(y_true, y_pred, y_true_ex, y_pred_ex):
    true, pred = ensure_inputs(y_true, y_pred, k=_K)

    # First check shapes (np does broadcasting for its testing functions)
    assert true.shape == y_true_ex.shape
    assert pred.shape == y_pred_ex.shape

    np.testing.assert_array_equal(true, y_true_ex)
    np.testing.assert_array_equal(pred, y_pred_ex)



import warnings
import numpy as np

from functools import wraps
from irmetrics.relevance import multilabel


def to_scalar(x):
    if not x.shape:
        return x.item()
    return x


def ensure_inputs(y_true, y_pred, k=None):
    y_true, y_pred = np.atleast_2d(y_true, y_pred)

    # `np.atleast_2d` adds a new axis as a batch dimension
    # thus y_pred[n_samples] is converted to y_pred[1, n_samples]
    # In other words, this condition allows passing y_true of [n_samples,]
    if y_true.shape[0] == 1 and y_true.shape[1] == y_pred.shape[0]:
        y_true = y_true.T

    # Take at most k labels
    y_true = y_true[:, :k]
    y_pred = y_pred[:, :k]
    return y_true, y_pred


def _validate_unique(f):
    @wraps(f)
    def wrapper(y_true, y_pred, k=None, relevance=multilabel, **kwargs):
        repeat_count = (y_pred[:, :, None] == y_pred[:, None]).sum(axis=-1)
        if np.any(repeat_count > 1):
            message = (
                "Repeated predictions detected. "
                "This is an error unless the predictions are padded. "
                "Use np.nan as a padding token to suppress the warning for "
                "integer labels."
            )
            warnings.warn(message, RuntimeWarning)
        return f(y_true, y_pred, k, relevance=relevance, **kwargs)

    return wrapper


def _ensure_io(f):
    @wraps(f)
    def wrapper(y_true, y_pred, k=None, relevance=multilabel, **kwargs):
        # Ensure (n_samples, n_labels) shapes for the inputs
        y_true_, y_pred_ = ensure_inputs(y_true, y_pred, k)

        # Calculate the measure
        raw_outputs = f(y_true_, y_pred_, k, relevance=relevance, **kwargs)

        # Remove unwanted dimensions if any
        return to_scalar(np.squeeze(raw_outputs))

    return wrapper



import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, "README.rst")).read()
except IOError:
    README = ""

setup(
    name="ir-metrics",
    use_scm_version=True,
    description="The most common information retrieval (IR) metrics",
    long_description=README,
    url="https://github.com/kqf/ir-metrics",
    setup_requires=["setuptools_scm"],
    packages=find_packages(),
    classifiers=[  # Optional
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
    ],
    python_requires=">=3.6, <4",
    install_requires=[
        "numpy",
    ],
    extras_require={
        # Didn't come up with a better name
        "pandas": ["pandas"],
    },
)



from functools import partial


def _relevance(y_true, y_pred):
    """
    This is a helper function to adjust convert the inputs for top-k metrics.

    Returns
    -------
    y_true : ndarray
        Passes through the this parameter unchanged.

    """
    return y_true


def flat(df, query_col, relevance_col, measure, k=None):
    """
    Calculate the corresponding measure for the data in flat format, with
    precalculated relevance judgements:

    +---------------+-------------------+-----------------+
    |   query_col   |   relevance_col   |   weights_col   |
    +---------------+-------------------+-----------------+
    |   1           |   0               |   1.0           |
    |   1           |   1               |   2.0           |
    |   1           |   0               |   3.0           |
    |   2           |   0               |   4.0           |
    |   2           |   1               |   5.0           |
    |   2           |   1               |   6.0           |
    |   2           |   1               |   7.0           |
    +---------------+-------------------+-----------------+

    Parameters
    ----------
    df : pandas.DataFrame
        Dataset in the flat form: each row corresponds to a sample with the
        given query_id and relevance judgement (higher is better).
    query_col :  str
        The column that corresponds to query identificator.
    relevance_col :  str
        The column that corresponds to relevance judgements.
    measure :  callable
        The desired measure to be calculated (one from `irmetrics.topk`).
        Currently, only ``topk.ndcg`` and ``topk.rr`` are supported.
    k : int, default=None
        Only consider the highest k scores in the ranking. If None, use all
        outputs.

    Returns
    -------
    measures : pandas.core.series.Series
        The values of the corresponding measure calculated per each query.

    Examples
    --------
    >>> import pandas as pd
    >>> from irmetrics.topk import rr
    >>> from irmetrics.flat import flat
    >>> df = pd.DataFrame({"quid": [1, 1, 2, 2], "rel": [1, 0, 0, 1]})
    >>> flat(df, query_col="quid", relevance_col="rel", measure=rr)
    quid
    1    1.0
    2    0.5
    Name: rel, dtype: float64
    """
    f = partial(measure, y_pred=None, k=k, relevance=_relevance)
    return df.groupby(query_col)[relevance_col].apply(f)
